# -*- coding: utf-8 -*-
"""
Created on Mon Nov 11 15:18:23 2024

@author: Giovanni
"""

import numpy as np
import matplotlib.pyplot as plt

def AAKR_reconstruction_ex(train_data_name, test_data_name, h):
    """
    Perform Adaptive Autoassociative Kernel Regression (AAKR) reconstruction.

    Parameters:
    train_data_name (str): Name of the file containing the training patterns.
    test_data_name (str): Name of the file containing the test patterns.
    h (float): Bandwidth parameter.

    Returns:
    test_data_rec (ndarray): Reconstruction of the test data.
    rmse (float): Root Mean Square Error indicating the similarity to measurements.
    """
    # Load train and test data
    train_data = np.loadtxt(train_data_name)
    test_data = np.loadtxt(test_data_name)

    # Get data dimensions
    N, n = train_data.shape
    N_test, _ = test_data.shape

    # Data normalization
    m_data = np.mean(train_data, axis=0)
    std_data = np.std(train_data, axis=0)
    train_data_n = (train_data - m_data) / std_data
    test_data_n = (test_data - m_data) / std_data

    # Initialize reconstructed data
    test_data_rec_n = np.zeros_like(test_data_n)

    # Perform reconstruction for each test pattern
    for ktest in range(N_test):
        # Compute distances and weights
        d2 = np.sum((test_data_n[ktest, :] - train_data_n) ** 2, axis=1)
        w = (1 / (np.sqrt(2 * np.pi) * h)) * np.exp(-d2 / (2 * h ** 2))
        
        # Handle case where sum of weights is zero
        if np.sum(w) == 0:
            w = np.ones(N) / N
        
        # Weighted reconstruction for each signal
        for isig in range(n):
            test_data_rec_n[ktest, isig] = np.sum(w * train_data_n[:, isig]) / np.sum(w)
    
    # Denormalize the reconstructed data
    test_data_rec = test_data_rec_n * std_data + m_data

    # Compute the root mean square error (RMSE)
    square_error=(test_data_rec - test_data) ** 2
    mean_square_error= np.sum(square_error,0) / N_test
    rmse_all = np.sqrt(mean_square_error)
    rmse = np.sum(rmse_all) / (n)


    # Optional plotting of the reconstruction and residuals
    signal_names = ['T1', 'T2', 'T3', 'T4', 'T5', 'T6']
    measure_error = 0.3
    for i in range(min(6, n)):
        plt.figure(i)
        
        # Plot original vs. reconstructed signals
        plt.subplot(2, 1, 1)
        plt.plot(test_data[:, i], 'b', label='Measurements')
        plt.plot(test_data_rec[:, i], 'r', label='Reconstruction')
        plt.xlabel(signal_names[i])
        plt.legend()
        
        # Plot residuals with upper and lower error bounds
        plt.subplot(2, 1, 2)
        residuals = -test_data_rec[:, i] + test_data[:, i]
        plt.plot(residuals, 'r', label='Residual')
        upper_error = np.ones(N_test) * 5 * measure_error
        lower_error = -upper_error
        plt.plot(upper_error, 'g--')
        plt.plot(lower_error, 'g--')
        plt.ylabel('Residual')
        plt.xlabel(signal_names[i])
        plt.ylim(-20 * measure_error, 20 * measure_error)
        plt.grid(True)
        plt.show()

    return test_data_rec, rmse
